# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

import pyrato  # noqa

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.imgmath',
    'autodocsumm',
    'sphinx_design',
    'sphinx_favicon',
]

# show tocs for classes and functions of modules using the autodocsumm
# package
autodoc_default_options = {'autosummary': True}

# show the code of plots that follows the command .. plot:: based on the
# package matplotlib.sphinxext.plot_directive
plot_include_source = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'pyrato'
copyright = u"2021-2023, Marco Berzborn; 2023, The pyfar developers"
author = u"The pyfar developers"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = pyrato.__version__
# The full version, including alpha/beta/rc tags.
release = pyrato.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'


# List of patterns, relative to source directory, that match files and
templates_path = ['_templates']
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path


exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use (Not defining
# uses the default style of the html_theme).
# pygments_style = 'sphinx'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# default language for highlighting in source code
highlight_language = "python3"

# intersphinx mapping
intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'spharpy': ('https://spharpy.readthedocs.io/en/stable/', None)
    }

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']
html_css_files = ['css/custom.css']
html_logo = 'resources/logos/pyfar_logos_fixed_size_pyrato.png'
html_title = "pyfar"
html_favicon = '_static/favicon.ico'


# -- HTML theme options
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html

html_theme_options = {
    "navbar_start": ["navbar-logo"],
    "navbar_end": ["navbar-icon-links"],
    "navbar_align": "content",
    "icon_links": [
        {
          "name": "GitHub",
          "url": "https://github.com/pyfar",
          "icon": "fa-brands fa-square-github",
          "type": "fontawesome",
        },
        {
            "name": "CC-BY",
            "url": "https://creativecommons.org/licenses/by/4.0/deed.de",
            "icon": "fa-brands fa-creative-commons-by",
            "type": "fontawesome",
        }
    ],

  # Configure secondary (right) side bar
  "show_toc_level": 3,                     # Show all subsections of notebooks
  "secondary_sidebar_items": ["page-toc"],  # Omit 'show source' link that that
                                           # shows notebook in json format
  "navigation_with_keys": True,
}

html_context = {
   "default_mode": "light"
}
# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# texinfo_documents = [
#     (master_doc, 'pyfar',
#      u'pyfar Gallery',
#      author,
#      'pyfar',
#      'One line description of project.',
#      'Miscellaneous'),
# ]

# -- Options for nbsphinx -------------------------------------------------
nbsphinx_prolog = r"""
{% set docname = 'doc/' + env.doc2path(env.docname, base=None) %}

.. raw:: html

    <div class="admonition note">
      Open an interactive online version by clicking the badge
      <span style="white-space: nowrap;"><a href="https://mybinder.org/v2/gh/pyfar/gallery/main?filepath={{ docname|e }}"><img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a></span>
      or
      <a href="{{ env.docname.split('/')|last|e + '.ipynb' }}" class="reference download internal" download>download</a>
      the notebook.
      <script>
        if (document.location.host) {
          let nbviewer_link = document.createElement('a');
          nbviewer_link.setAttribute('href',
            'https://nbviewer.org/url' +
            (window.location.protocol == 'https:' ? 's/' : '/') +
            window.location.host +
            window.location.pathname.slice(0, -4) +
            'ipynb');
          nbviewer_link.innerHTML = 'Or view it on <em>nbviewer</em>';
          nbviewer_link.classList.add('reference');
          nbviewer_link.classList.add('external');
          document.currentScript.replaceWith(nbviewer_link, '.');
        }
      </script>
    </div>

"""

# -- manage thumbnails --------------------------------------------------------
# must be located in 'docs/_static'
nbsphinx_thumbnails = {
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The edc_noise_handling module provides various methods for noise
    compensation of room impulse responses.
"""

import numpy as np
from matplotlib import pyplot as plt
from pyrato import dsp
import warnings
import pyfar as pf


def subtract_noise_from_squared_rir(data, noise_level='auto'):
    """Subtract the noise power from a squared room impulse response.

    Parameters
    ----------
    data : pyfar.TimeData
        The squared room impulse response.
    noise_level : str or numpy.ndarray, float, optional
        The noise power for each channel. The default is 'auto', which will
        try to estimate the noise power from the room impulse response.

    Returns
    -------
    pyfar.TimeData
        The squared room impulse response after noise power subtraction.

    """
    subtracted = _subtract_noise_from_squared_rir(
        data.time, noise_level=noise_level)

    return pf.TimeData(subtracted, data.times, comment=data.comment)


def _subtract_noise_from_squared_rir(data, noise_level='auto'):
    """Subtract the noise power from a squared room impulse response.

    Parameters
    ----------
    data : ndarray, double
        The squared room impulse response with dimension [..., n_samples]

    Returns
    -------
    result : ndarray, double
        The noise-subtracted RIR.

    """
    if noise_level == "auto":
        noise_level = dsp._estimate_noise_energy(
            data,
            interval=[0.9, 1.0])
    return (data.T - noise_level).T


def schroeder_integration(room_impulse_response, is_energy=False):
    r"""Calculate the Schroeder integral of a room impulse response [#]_. The
    result is the energy decay curve for the given room impulse response.

    .. math:

        \langle e^2(t) \rangle = N\cdot \int_{t}^{\infty} h^2(\tau)
        \mathrm{d} \tau

    Parameters
    ----------
    room_impulse_response : pyfar.Signal
        Room impulse response as array
    is_energy : boolean, optional
        Whether the input represents energy data or sound pressure values.

    Returns
    -------
    energy_decay_curve : pyfar.TimeData
        The energy decay curve

    Note
    ----
    This function does not apply any compensation of measurement noise and
    integrates the full length of the input signal. It should only be used
    if no measurement noise or artifacts are present in the data.

    References
    ----------
    .. [#]  M. R. Schroeder, “New Method of Measuring Reverberation Time,”
            The Journal of the Acoustical Society of America, vol. 37, no. 6,
            pp. 1187-1187, 1965.

    Example
    -------
    Calculate the Schroeder integral of a simulated RIR and plot.

    .. plot::

        >>> import numpy as np
        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> from pyrato.analytic import rectangular_room_rigid_walls
        ...
        >>> L = np.array([8, 5, 3])/10
        >>> source_pos = np.array([5, 3, 1.2])/10
        >>> receiver_pos = np.array([1, 1, 1.2])/10
        >>> rir, _ = rectangular_room_rigid_walls(
        ...     L, source_pos, receiver_pos,
        ...     reverberation_time=1, max_freq=1e3, n_samples=2**16,
        ...     speed_of_sound=343.9)
        >>> edc = ra.schroeder_integration(rir)
        >>> pf.plot.time(rir/np.abs(rir.time).max(), dB=True, label='RIR')
        >>> ax = pf.plot.time(
        ...     edc/edc.time[..., 0], dB=True, log_prefix=10, label='EDC')
        >>> ax.set_ylim(-65, 5)
        >>> ax.legend()
    """

    edc = _schroeder_integration(
        room_impulse_response.time, is_energy=is_energy)

    return pf.TimeData(edc, room_impulse_response.times)


def _schroeder_integration(impulse_response, is_energy=False):
    r"""Calculate the Schroeder integral of a room impulse response [#]_. The
    result is the energy decay curve for the given room impulse response.

    .. math:

        \langle e^2(t) \rangle = N\cdot \int_{t}^{\infty} h^2(\tau)
        \mathrm{d} \tau

    Parameters
    ----------
    impulse_response : ndarray, double
        Room impulse response as array
    is_energy : boolean, optional
        Whether the input represents energy data or sound pressure values.

    Returns
    -------
    energy_decay_curve : ndarray, double
        The energy decay curve

    References
    ----------
    .. [#]  M. R. Schroeder, “New Method of Measuring Reverberation Time,”
            The Journal of the Acoustical Society of America, vol. 37, no. 6,
            pp. 1187-1187, 1965.

    """
    if not is_energy:
        data = np.abs(impulse_response)**2
    else:
        data = impulse_response.copy()

    ndim = data.ndim
    data = np.atleast_2d(data)
    energy_decay_curve = np.fliplr(np.nancumsum(np.fliplr(data), axis=-1))

    if ndim < energy_decay_curve.ndim:
        energy_decay_curve = np.squeeze(energy_decay_curve)

    return energy_decay_curve


def energy_decay_curve_truncation(
        data,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=15,
        plot=False):
    """ This function truncates a given room impulse response by the
    intersection time after Lundeby and calculates the energy decay curve.

    Parameters
    ----------
    data : pyfar.Signal
        The room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalization is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    threshold : float
        Defines a peak-signal-to-noise ratio based threshold in dB for final
        truncation of the EDC. Values below the sum of the threshold level and
        the peak-signal-to-noise ratio in dB are discarded. The default is
        15 dB, which is in correspondence with ISO 3382-1:2009 [#]_.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    pyfar.TimeData
        Returns the noise compensated EDC.

    Examples
    --------

    Plot the RIR and the EDC calculated truncating the integration at the
    intersection time.

    .. plot::

        >>> import numpy as np
        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> from pyrato.analytic import rectangular_room_rigid_walls
        ...
        >>> L = np.array([8, 5, 3])/10
        >>> source_pos = np.array([5, 3, 1.2])/10
        >>> receiver_pos = np.array([1, 1, 1.2])/10
        >>> rir, _ = rectangular_room_rigid_walls(
        ...     L, source_pos, receiver_pos,
        ...     reverberation_time=1, max_freq=1e3, n_samples=2**16,
        ...     speed_of_sound=343.9)
        >>> rir = rir/rir.time.max()
        ...
        >>> awgn = pf.signals.noise(
        ...     rir.n_samples, rms=rir.time.max()*10**(-50/20),
        ...     sampling_rate=rir.sampling_rate)
        >>> rir = rir + awgn
        >>> edc = ra.energy_decay_curve_truncation(rir)
        ...
        >>> ax = pf.plot.time(rir, dB=True, label='RIR')
        >>> pf.plot.time(edc, dB=True, log_prefix=10, label='EDC')
        >>> ax.set_ylim(-65, 5)
        >>> ax.legend()


    References
    ----------
    .. [#]  International Organization for Standardization, “EN ISO 3382-1:2009
            Acoustics - Measurement of room acoustic parameters,” 2009.

    """
    energy_data = dsp.preprocess_rir(
        data,
        is_energy=is_energy,
        shift=time_shift,
        channel_independent=channel_independent)
    n_samples = data.n_samples

    intersection_time = intersection_time_lundeby(
        energy_data,
        freq=freq,
        initial_noise_power=noise_level,
        is_energy=True,
        time_shift=False,
        channel_independent=False,
        plot=False)[0]

    intersection_time_idx = np.rint(intersection_time * data.sampling_rate)

    psnr = dsp.peak_signal_to_noise_ratio(
        data, noise_level, is_energy=is_energy)
    trunc_levels = 10*np.log10((psnr)) - threshold

    energy_decay_curve = np.zeros([*data.cshape, n_samples])
    for ch in np.ndindex(data.cshape):
        energy_decay_curve[
            ch, :int(intersection_time_idx[ch])] = \
                _schroeder_integration(
                    energy_data.time[
                        ch, :int(intersection_time_idx[ch])],
                    is_energy=True)

    energy_decay_curve = _truncate_energy_decay_curve(
        energy_decay_curve, trunc_levels)

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (energy_decay_curve.T /
                                  energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of each channel.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    edc = pf.TimeData(
        energy_decay_curve, data.times, comment=data.comment)

    if plot:
        ax = pf.plot.time(data, dB=True, label='RIR')
        pf.plot.time(edc, dB=True, log_prefix=10, label='EDC')
        ax.set_ylim(-65, 5)
        ax.legend()

    return edc


def energy_decay_curve_lundeby(
        data,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False):
    """Lundeby et al. [#]_ proposed a correction term to prevent the truncation
    error. The missing signal energy from truncation time to infinity is
    estimated and added to the truncated integral.

    Parameters
    ----------
    data : pyfar.Signal
        The room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    pyfar.TimeData
        Returns the noise handeled edc.

    References
    ----------
    .. [#]  Lundeby, Virgran, Bietz and Vorlaender - Uncertainties of
            Measurements in Room Acoustics - ACUSTICA Vol. 81 (1995)

    Examples
    --------

    Plot the RIR and the EDC calculated after Lundeby.

    .. plot::

        >>> import numpy as np
        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> from pyrato.analytic import rectangular_room_rigid_walls
        ...
        >>> L = np.array([8, 5, 3])/10
        >>> source_pos = np.array([5, 3, 1.2])/10
        >>> receiver_pos = np.array([1, 1, 1.2])/10
        >>> rir, _ = rectangular_room_rigid_walls(
        ...     L, source_pos, receiver_pos,
        ...     reverberation_time=1, max_freq=1e3, n_samples=2**16,
        ...     speed_of_sound=343.9)
        >>> rir = rir/rir.time.max()
        ...
        >>> awgn = pf.signals.noise(
        ...     rir.n_samples, rms=rir.time.max()*10**(-50/20),
        ...     sampling_rate=rir.sampling_rate)
        >>> rir = rir + awgn
        >>> edc = ra.energy_decay_curve_lundeby(rir)
        ...
        >>> ax = pf.plot.time(rir, dB=True, label='RIR')
        >>> pf.plot.time(edc, dB=True, log_prefix=10, label='EDC')
        >>> ax.set_ylim(-65, 5)
        >>> ax.legend()

    """

    energy_data = dsp.preprocess_rir(
        data,
        is_energy=is_energy,
        shift=time_shift,
        channel_independent=channel_independent)
    n_samples = energy_data.n_samples
    sampling_rate = data.sampling_rate

    intersection_time, late_reverberation_time, noise_estimation = \
        intersection_time_lundeby(
            energy_data,
            freq=freq,
            initial_noise_power=noise_level,
            is_energy=True,
            time_shift=False,
            channel_independent=False,
            plot=False)
    time_vector = data.times

    energy_decay_curve = np.zeros([*data.cshape, n_samples])

    for ch in np.ndindex(data.cshape):
        intersection_time_idx = np.argmin(
            np.abs(time_vector - intersection_time[ch]))
        p_square_at_intersection = noise_estimation[ch]

        # Calculate correction term according to DIN EN ISO 3382
        # TO-DO: check reference!
        correction = (p_square_at_intersection
                      * late_reverberation_time[ch]
                      * (1 / (6*np.log(10)))
                      * sampling_rate)

        energy_decay_curve[ch, :intersection_time_idx] = \
            _schroeder_integration(
                energy_data.time[ch, :intersection_time_idx],
                is_energy=True)
        energy_decay_curve[ch] += correction
        energy_decay_curve[ch, intersection_time_idx:] = np.nan

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (
                energy_decay_curve.T /
                energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of each channel.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    edc = pf.TimeData(
        energy_decay_curve, data.times, comment=data.comment)

    if plot:
        ax = pf.plot.time(data, dB=True, label='RIR')
        pf.plot.time(edc, dB=True, log_prefix=10, label='EDC')
        ax.set_ylim(-65, 5)
        ax.legend()

    return edc


def energy_decay_curve_chu(
        data,
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=10,
        plot=False):
    """ Implementation of the "subtraction of noise"-method after Chu [#]
    The noise level is estimated and subtracted from the impulse response
    before backward integration.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    threshold : float, None
        Defines a peak-signal-to-noise ratio based threshold in dB for final
        truncation of the EDC. Values below the sum of the threshold level and
        the peak-signal-to-noise ratio in dB are discarded. The default is
        10 dB. If `None`, the decay curve will not be truncated further.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    energy_decay_curve: ndarray, double
        Returns the noise handeled edc.

    References
    ----------
    .. [#]  W. T. Chu. “Comparison of reverberation measurements using
            Schroeder’s impulse method and decay-curve averaging method”.
            In: Journal of the Acoustical Society of America 63.5 (1978),
            pp. 1444–1450.

    Examples
    --------

    .. plot::

        >>> import numpy as np
        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> from pyrato.analytic import rectangular_room_rigid_walls
        ...
        >>> L = np.array([8, 5, 3])/10
        >>> source_pos = np.array([5, 3, 1.2])/10
        >>> receiver_pos = np.array([1, 1, 1.2])/10
        >>> rir, _ = rectangular_room_rigid_walls(
        ...     L, source_pos, receiver_pos,
        ...     reverberation_time=1, max_freq=1e3, n_samples=2**16,
        ...     speed_of_sound=343.9)
        ...
        >>> awgn = pf.signals.noise(
        ...     rir.n_samples, rms=rir.time.max()*10**(-40/20),
        ...     sampling_rate=rir.sampling_rate)
        >>> rir = rir + awgn
        >>> edc = ra.energy_decay_curve_chu(rir)
        ...
        >>> pf.plot.time(rir/np.abs(rir.time).max(), dB=True, label='RIR')
        >>> ax = pf.plot.time(
        ...     edc/edc.time[..., 0], dB=True, log_prefix=10, label='EDC')
        >>> ax.set_ylim(-65, 5)
        >>> ax.legend()

    """
    energy_data = dsp.preprocess_rir(
        data,
        is_energy=is_energy,
        shift=time_shift,
        channel_independent=channel_independent)

    subtracted = subtract_noise_from_squared_rir(
        energy_data,
        noise_level=noise_level)

    edc = schroeder_integration(subtracted, is_energy=True)

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            edc.time = (edc.time.T / edc.time[..., 0]).T
        else:
            # ...by the maximum first element of all channels.
            max_start_value = np.amax(edc.time[..., 0])
            edc.time /= max_start_value

    mask = edc.time <= 2*np.finfo(float).eps
    if np.any(mask):
        first_zero = np.nanargmax(mask, axis=-1)
        for ch in np.ndindex(edc.cshape):
            edc.time[ch, first_zero[ch]:] = np.nan

    if threshold is not None:
        psnr = dsp.peak_signal_to_noise_ratio(
            data, noise_level, is_energy=is_energy)
        trunc_levels = 10*np.log10((psnr)) - threshold
        edc = truncate_energy_decay_curve(edc, trunc_levels)

    if plot:
        plt.figure(figsize=(15, 3))
        pf.plot.use('light')
        plt.subplot(131)
        pf.plot.time(energy_data, dB=True, log_prefix=10)
        plt.ylabel('Squared IR in dB')
        plt.subplot(132)
        pf.plot.time(subtracted, dB=True, log_prefix=10)
        plt.ylabel('Noise subtracted RIR in dB')
        plt.subplot(133)
        pf.plot.time(edc, dB=True, log_prefix=10)
        plt.ylabel('EDC in dB')

    return edc


def energy_decay_curve_chu_lundeby(
        data,
        freq='broadband',
        noise_level='auto',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False):
    """ This function combines Chu's and Lundeby's methods:
    The estimated noise level is subtracted before backward integration,
    the impulse response is truncated at the intersection time,
    and the correction for the truncation is applied [4]_, [5]_, [6]_

    Parameters
    ----------
    data : pyfar.Signal
        The room impulse response.
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    normalize : boolean
        Defines, if the energy decay curve should be normalized in the end
        or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    pyfar.TimeData
        Returns the noise handeled edc.

    References
    ----------
    .. [4]  Lundeby, Virgran, Bietz and Vorlaender - Uncertainties of
            Measurements in Room Acoustics - ACUSTICA Vol. 81 (1995)
    .. [5]  W. T. Chu. “Comparison of reverberation measurements using
            Schroeder’s impulse method and decay-curve averaging method”. In:
            Journal of the Acoustical Society of America 63.5 (1978),
            pp. 1444–1450.
    .. [6]  M. Guski, “Influences of external error sources on measurements of
            room acoustic parameters,” 2015.

    Examples
    --------

    Calculate and plot the EDC using a combination of Chu's and Lundeby's
    methods.

    .. plot::

        >>> import numpy as np
        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> from pyrato.analytic import rectangular_room_rigid_walls
        ...
        >>> L = np.array([8, 5, 3])/10
        >>> source_pos = np.array([5, 3, 1.2])/10
        >>> receiver_pos = np.array([1, 1, 1.2])/10
        >>> rir, _ = rectangular_room_rigid_walls(
        ...     L, source_pos, receiver_pos,
        ...     reverberation_time=1, max_freq=1e3, n_samples=2**16,
        ...     speed_of_sound=343.9)
        >>> rir = rir/rir.time.max()
        ...
        >>> awgn = pf.signals.noise(
        ...     rir.n_samples, rms=rir.time.max()*10**(-50/20),
        ...     sampling_rate=rir.sampling_rate)
        >>> rir = rir + awgn
        >>> edc = ra.energy_decay_curve_chu_lundeby(rir)
        ...
        >>> ax = pf.plot.time(rir, dB=True, label='RIR')
        >>> pf.plot.time(edc, dB=True, log_prefix=10, label='EDC')
        >>> ax.set_ylim(-65, 5)
        >>> ax.legend()

    """

    energy_data = dsp.preprocess_rir(
        data,
        is_energy=is_energy,
        shift=time_shift,
        channel_independent=channel_independent)
    n_samples = energy_data.n_samples

    subtraction = subtract_noise_from_squared_rir(
        energy_data,
        noise_level=noise_level)

    intersection_time, late_reverberation_time, noise_level = \
        intersection_time_lundeby(
            energy_data,
            freq=freq,
            initial_noise_power=noise_level,
            is_energy=True,
            time_shift=False,
            channel_independent=False,
            plot=False)

    time_vector = data.times
    energy_decay_curve = np.zeros([*data.cshape, n_samples])

    for ch in np.ndindex(data.cshape):
        intersection_time_idx = np.argmin(np.abs(
            time_vector - intersection_time[ch]))
        if type(noise_level) is str and noise_level == 'auto':
            p_square_at_intersection = dsp.estimate_noise_energy(
                energy_data.time[ch], is_energy=True)
        else:
            p_square_at_intersection = noise_level[ch]

        # calculate correction term according to DIN EN ISO 3382
        correction = (p_square_at_intersection
                      * late_reverberation_time[ch]
                      * (1 / (6*np.log(10)))
                      * data.sampling_rate)

        energy_decay_curve[ch, :intersection_time_idx] = \
            _schroeder_integration(
                subtraction.time[ch, :intersection_time_idx],
                is_energy=True)
        energy_decay_curve[ch] += correction
        energy_decay_curve[ch, intersection_time_idx:] = np.nan

    if normalize:
        # Normalize the EDC...
        if not channel_independent:
            # ...by the first element of each channel.
            energy_decay_curve = (energy_decay_curve.T /
                                  energy_decay_curve[..., 0]).T
        else:
            # ...by the maximum first element of each channel.
            max_start_value = np.amax(energy_decay_curve[..., 0])
            energy_decay_curve /= max_start_value

    edc = pf.TimeData(
        energy_decay_curve, data.times, comment=data.comment)

    if plot:
        ax = pf.plot.time(data, dB=True, label='RIR')
        pf.plot.time(edc, dB=True, log_prefix=10, label='EDC')
        ax.set_ylim(-65, 5)
        ax.legend()

    return edc


def intersection_time_lundeby(
        data,
        freq='broadband',
        initial_noise_power='auto',
        is_energy=False,
        time_shift=False,
        channel_independent=False,
        plot=False):
    """Calculate the intersection time between impulse response and noise.

    This function uses the algorithm after Lundeby et al. [#]_ to calculate
    the intersection time, lundeby reverberation time, and noise level
    estimation.

    Parameters
    ----------
    data : pyfar.Signal
        The room impulse response
    freq: integer OR string
        The frequency band. If set to 'broadband',
        the time window of the Lundeby-algorithm will not be set in dependence
        of frequency.
    noise_level: ndarray, double OR string
        If not specified, the noise level is calculated based on the last 10
        percent of the RIR. Otherwise specify manually for each channel
        as array.
    is_energy: boolean
        Defines, if the data is already squared.
    time_shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift and normalizsation is done
        channel-independently or not.
    plot: Boolean
        Specifies, whether the results should be visualized or not.

    Returns
    -------
    intersection_time: ndarray, float
        Returns the Lundeby intersection time for each channel.
    reverberation_time: ndarray, float
        Returns the Lundeby reverberation time for each channel.
    noise_level: ndarray, float
        Returns the noise level estimation for each channel.

    References
    ----------
    .. [#]  Lundeby, Virgran, Bietz and Vorlaender - Uncertainties of
            Measurements in Room Acoustics - ACUSTICA Vol. 81 (1995)

    Examples
    --------

    Estimate the intersection time :math:`T_i` and plot the RIR and the
    estimated noise power.

    .. plot::

        >>> import numpy as np
        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> from pyrato.analytic import rectangular_room_rigid_walls
        ...
        >>> L = np.array([8, 5, 3])/10
        >>> source_pos = np.array([5, 3, 1.2])/10
        >>> receiver_pos = np.array([1, 1, 1.2])/10
        >>> rir, _ = rectangular_room_rigid_walls(
        ...     L, source_pos, receiver_pos,
        ...     reverberation_time=1, max_freq=1e3, n_samples=2**16,
        ...     speed_of_sound=343.9)
        >>> rir = rir/np.abs(rir.time).max()
        ...
        >>> awgn = pf.signals.noise(
        ...     rir.n_samples, rms=rir.time.max()*10**(-40/20),
        ...     sampling_rate=rir.sampling_rate)
        >>> rir = rir + awgn
        >>> inter_time, _, noise_power = ra.intersection_time_lundeby(rir)
        ...
        >>> ax = pf.plot.time(rir, dB=True, label='RIR')
        >>> ax.axvline(inter_time, c='k', linestyle='--', label='$T_i$')
        >>> ax.axhline(
        ...     10*np.log10(noise_power), c='k', linestyle=':', label='Noise')
        >>> ax.set_ylim(-65, 5)
        >>> ax.legend()

    """
    # Define constants:
    # time intervals per 10 dB decay. Lundeby: 3...10
    n_intervals_per_10dB = 5
    # end of regression 5 ... 10 dB
    dB_above_noise = 10
    # Dynamic range 10 ... 20 dB
    use_dyn_range_for_regression = 20

    energy_data = dsp.preprocess_rir(
        data,
        is_energy=is_energy,
        shift=time_shift,
        channel_independent=channel_independent)

    if isinstance(data, pf.Signal):
        sampling_rate = data.sampling_rate
    elif isinstance(data, pf.TimeData) and not isinstance(data, pf.Signal):
        sampling_rate = np.round(1/np.diff(data.times).mean(), decimals=4)
    energy_data = energy_data.time

    if freq == "broadband":
        # broadband: use 30 ms windows sizes
        freq_dependent_window_time = 0.03
    else:
        freq_dependent_window_time = (800/freq+10) / 1000

    # (1) SMOOTH
    time_window_data, time_vector_window, time_vector = dsp._smooth_rir(
        energy_data, sampling_rate, freq_dependent_window_time)

    # (2) ESTIMATE NOISE
    if initial_noise_power == 'auto':
        noise_estimation = dsp._estimate_noise_energy(energy_data)
    else:
        noise_estimation = initial_noise_power.copy()

    # (3) REGRESSION
    reverberation_time = np.zeros(data.cshape, data.time.dtype)
    noise_level = np.zeros(data.cshape, data.time.dtype)
    intersection_time = np.zeros(data.cshape, data.time.dtype)
    noise_peak_level = np.zeros(data.cshape, data.time.dtype)

    for ch in np.ndindex(data.cshape):
        time_window_data_current_channel = time_window_data[ch]
        start_idx = np.nanargmax(time_window_data_current_channel, axis=-1)
        try:
            stop_idx = (np.argwhere(10*np.log10(
                time_window_data_current_channel[start_idx+1:-1]) >
                    (10*np.log10(noise_estimation[ch]) +
                        dB_above_noise))[-1, 0] + start_idx)
        except IndexError:
            raise Exception(
                'Regression failed: Low SNR. Estimation terminated.')

        dyn_range = np.diff(10*np.log10(np.take(
            time_window_data_current_channel, [start_idx, stop_idx])))

        if (stop_idx == start_idx) or dyn_range > -5:
            raise Exception(
                'Regression failed: Low SNR. Estimation terminated.')

        # regression_matrix*slope = edc
        regression_matrix = np.vstack((np.ones(
            [stop_idx-start_idx]), time_vector_window[start_idx:stop_idx]))
        slope = np.linalg.lstsq(
            regression_matrix.T,
            10*np.log10(time_window_data_current_channel[start_idx:stop_idx]),
            rcond=None)[0]

        if slope[1] == 0 or np.any(np.isnan(slope)):
            raise Exception(
                'Regression failed: T would be Inf, setting to 0. \
                    Estimation terminated.')

        regression_time = np.array(
            [time_vector_window[start_idx], time_vector_window[stop_idx]])
        regression_values = np.array(
            [10*np.log10(time_window_data[0, start_idx]),
             (10*np.log10(time_window_data[0, start_idx])
                + slope[1]*time_vector_window[stop_idx])])

        # (4) PRELIMINARY CROSSING POINT
        crossing_point = \
            (10*np.log10(noise_estimation[ch]) - slope[0]) / slope[1]
        preliminary_crossing_point = crossing_point

        # (5) NEW LOCAL TIME INTERVAL LENGTH
        n_blocks_in_decay = (np.diff(
            10*np.log10(np.take(
                time_window_data_current_channel, [start_idx, stop_idx])))
            / -10 * n_intervals_per_10dB)

        n_samples_per_block = np.round(np.diff(np.take(
            time_vector_window,
            [start_idx, stop_idx])) / n_blocks_in_decay * sampling_rate)

        window_time = n_samples_per_block/sampling_rate

        # (6) AVERAGE
        time_window_data_current_channel, \
            time_vector_window_current_channel, \
            time_vector_current_channel = dsp._smooth_rir(
                energy_data[ch], sampling_rate, window_time)
        time_window_data_current_channel = np.squeeze(
            time_window_data_current_channel)
        idx_max = np.nanargmax(time_window_data_current_channel)

        # high start value to enter while-loop
        old_crossing_point = 11+crossing_point
        loop_counter = 0

        while True:
            # (7) ESTIMATE BACKGROUND LEVEL
            corresponding_decay = 10  # 5...10 dB
            idx_last_10_percent = np.round(
                time_window_data_current_channel.shape[-1]*0.9)

            t_block = n_samples_per_block / sampling_rate
            rel_decay = corresponding_decay / slope[1]
            idx_10dB_below_crosspoint = np.nanmax(
                np.r_[1, np.round(((crossing_point - rel_decay) / t_block))])

            noise_estimation_current_channel = np.nanmean(
                time_window_data_current_channel[int(np.nanmin(
                    [idx_last_10_percent, idx_10dB_below_crosspoint])):])

            # (8) ESTIMATE LATE DECAY SLOPE
            try:
                start_idx_loop = np.argwhere(10*np.log10(
                    time_window_data_current_channel[idx_max:]) < (
                        10*np.log10(noise_estimation_current_channel)
                        + dB_above_noise
                        + use_dyn_range_for_regression
                        + idx_max))[0, 0]
            except TypeError:
                start_idx_loop = 0

            try:
                stop_idx_loop = (np.argwhere(10*np.log10(
                    time_window_data_current_channel[start_idx_loop+1:]) < (
                        10*np.log10(noise_estimation_current_channel)
                        + dB_above_noise))[0, 0]
                                 + start_idx_loop)
            except IndexError:
                raise Exception(
                    'Regression failed: Low SNR. Estimation terminated.')

            # regression_matrix*slope = edc
            regression_matrix = np.vstack((np.ones(
                [stop_idx_loop-start_idx_loop]),
                time_vector_window_current_channel[
                    start_idx_loop:stop_idx_loop]))

            slope = np.linalg.lstsq(
                regression_matrix.T,
                (10*np.log10(time_window_data_current_channel[
                    start_idx_loop:stop_idx_loop])),
                rcond=None)[0]

            if slope[1] >= 0:
                raise Exception(
                    'Regression did not work due, T would be Inf, \
                        setting to 0. Estimation was terminated.')

            # (9) FIND CROSSPOINT
            old_crossing_point = crossing_point
            crossing_point = ((10*np.log10(noise_estimation_current_channel)
                               - slope[0]) / slope[1])

            loop_counter = loop_counter + 1

            if (np.abs(old_crossing_point-crossing_point) < 0.01):
                break
            if loop_counter > 30:
                # TO-DO: Paper says 5 iterations are sufficient in all cases!
                warnings.warn(
                    "Lundeby algorithm was terminated after 30 iterations.")
                break

        reverberation_time[ch] = -60/slope[1]
        noise_level[ch] = noise_estimation_current_channel
        intersection_time[ch] = crossing_point
        noise_peak_level[ch] = 10 * np.log10(
            np.nanmax(time_window_data_current_channel[int(np.nanmin(
                [idx_last_10_percent, idx_10dB_below_crosspoint])):]))

    if plot:
        plt.figure(figsize=(15, 3))
        plt.subplot(131)
        max_data_db = np.nanmax(10*np.log10(energy_data))
        plt.plot(time_vector, 10*np.log10(energy_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Squared RIR [dB]')
        plt.ylim(bottom=max_data_db-80+5, top=max_data_db+5)
        plt.grid(True)

        plt.subplot(132)
        plt.plot(time_vector_window, 10*np.log10(time_window_data.T))
        plt.xlabel('Time [s]')
        plt.ylabel('Smoothened RIR [dB]')
        plt.plot(
            time_vector_window[-int(np.round(time_window_data.shape[-1]/10))],
            10*np.log10(noise_estimation[0]),
            marker='o',
            label='noise estimation',
            linestyle='--',
            color='C1')
        plt.axhline(
                10*np.log10(noise_estimation[0]),
                color='C1',
                linestyle='--')
        plt.plot(
            regression_time,
            regression_values,
            marker='o',
            color='C2',
            label='regression')
        plt.plot(
            preliminary_crossing_point,
            10*np.log10(noise_estimation[ch]),
            marker='o',
            color='C3',
            label='preliminary crosspoint')
        plt.legend()
        plt.grid(True)

        plt.subplot(133)
        # TO-DO: plot all channels?
        plt.plot(
            time_vector_window_current_channel,
            10*np.log10(time_window_data_current_channel))
        plt.xlabel('Time [s]')
        plt.ylabel('Final EDC Estimation [dB]')
        plt.plot(
            time_vector_window_current_channel[int(idx_10dB_below_crosspoint)],
            10*np.log10(noise_estimation[0]),
            marker='o',
            color='C1',
            label='Noise')
        plt.axhline(
            10*np.log10(noise_estimation[0]),
            color='C1',
            linestyle='--')
        plt.plot(
            crossing_point,
            10*np.log10(noise_estimation[0]),
            marker='o',
            color='C3',
            label='Final crosspoint',
            linestyle='--')
        plt.axvline(
            reverberation_time[0],
            label=('$T_{60} = '+str(np.round(reverberation_time[0], 3)) + '$'),
            color='k')
        plt.axvline(
            intersection_time[0],
            label=('Lundeby intersection time'),
            color='C3',
            linestyle='--')
        plt.tight_layout()
        plt.legend()
        plt.grid(True)

    return intersection_time, reverberation_time, noise_level


def _truncate_energy_decay_curve(energy_decay_curve, threshold_level):

    edc = np.atleast_2d(energy_decay_curve)
    threshold_level = np.atleast_2d(threshold_level)
    e = edc.T[0]
    t = e / 10**(threshold_level/10)

    mask = edc.T < np.broadcast_to(t, edc.T.shape)
    edc[mask.T] = np.nan

    return edc


def truncate_energy_decay_curve(energy_decay_curve, threshold):
    """Truncate an energy decay curve, discarding values below the threshold.

    Parameters
    ----------
    energy_decay_curve : pyfar.TimeData
        The energy decay curve
    threshold : float
        The threshold level in dB. The data below the threshold level are set
        to numpy.nan values.
    """
    return pf.TimeData(
        _truncate_energy_decay_curve(energy_decay_curve.time, threshold),
        energy_decay_curve.times,
        energy_decay_curve.comment)



# -*- coding: utf-8 -*-

import os

import numpy as np
import numpy.testing as npt
import pytest
import pyfar as pf
from numpy import genfromtxt
import pyrato.dsp as dsp

test_data_path = os.path.join(os.path.dirname(__file__), 'test_data')


def mock_shift_samples_1d(*args, **kwargs):
    return np.array([76])


def mock_shift_samples_2d(*args, **kwargs):
    return np.array([76, 76])


def test_start_ir_insufficient_snr():
    n_samples = 2**9
    ir = np.zeros(n_samples, dtype=float)
    ir[20] = 1
    ir = pf.Signal(ir, 44100)

    snr = 15

    noise = np.random.randn(n_samples)
    noise = noise / np.sqrt(np.mean(np.abs(noise**2))) * 10**(-snr/20)
    noise = pf.Signal(noise, 44100)

    ir_noise = ir + noise

    with pytest.warns(match='SNR seems lower'):
        dsp.find_impulse_response_start(ir_noise)


def test_start_ir():
    n_samples = 2**10
    ir = np.zeros(n_samples)
    snr = 60

    noise = pf.Signal(np.random.randn(n_samples) * 10**(-snr/20), 44100)

    start_sample = 24
    ir[start_sample] = 1

    ir = pf.Signal(ir, 44100)

    start_sample_est = dsp.find_impulse_response_start(ir)
    assert start_sample_est == start_sample - 1

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_start(ir_awgn)
    assert start_sample_est == start_sample - 1


def test_start_ir_thresh():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    start_sample = 24
    ir[start_sample] = 1
    ir[start_sample-4:start_sample] = 10**(-5/10)

    ir = pf.Signal(ir, 44100)

    start_sample_est = dsp.find_impulse_response_start(ir, threshold=20)
    assert start_sample_est == start_sample - 4 - 1


def test_start_ir_multidim():
    n_samples = 2**10
    n_channels = 3
    ir = np.zeros((n_channels, n_samples))

    snr = 60

    noise = pf.Signal(
        np.random.randn(n_channels, n_samples) * 10**(-snr/20), 44100)

    start_sample = [24, 5, 43]
    ir[[0, 1, 2], start_sample] = 1

    ir = pf.Signal(ir, 44100)

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_start(ir_awgn)

    npt.assert_allclose(start_sample_est, np.array(start_sample) - 1)

    ir = np.zeros((2, n_channels, n_samples))
    noise = pf.Signal(
        np.random.randn(2, n_channels, n_samples) * 10**(-snr/20), 44100)

    start_sample_1 = [24, 5, 43]
    ir[0, [0, 1, 2], start_sample_1] = 1
    start_sample_2 = [14, 12, 16]
    ir[1, [0, 1, 2], start_sample_2] = 1

    ir = pf.Signal(ir, 44100)

    start_samples = np.vstack((start_sample_1, start_sample_2))

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_start(ir_awgn)

    npt.assert_allclose(start_sample_est, start_samples - 1)


def test_start_room_impulse_response():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')

    rir = pf.Signal(rir, 44100)

    actual = dsp.find_impulse_response_start(rir, threshold=20)

    expected = 0

    npt.assert_allclose(actual, expected)


def test_start_room_impulse_response_shfted(monkeypatch):
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')

    rir_shifted = np.roll(rir, 128, axis=-1)
    rir_shifted = pf.Signal(rir_shifted, 44100)
    actual = dsp.find_impulse_response_start(rir_shifted, threshold=20)

    expected = 128

    npt.assert_allclose(actual, expected)


def test_max_ir():
    n_samples = 2**10
    ir = np.zeros(n_samples)

    snr = 60

    noise = pf.Signal(
        np.random.randn(n_samples) * 10**(-snr/20), 44100)

    start_sample = 24
    ir[start_sample] = 1
    ir = pf.Signal(ir, 44100)
    start_sample_est = dsp.find_impulse_response_maximum(ir)
    assert start_sample_est == start_sample

    ir_awgn = ir + noise
    start_sample_est = dsp.find_impulse_response_maximum(ir_awgn)
    assert start_sample_est == start_sample

    with pytest.warns(match='SNR seems lower'):
        start_sample_est = dsp.find_impulse_response_maximum(
            ir_awgn, threshold=200)

# ------------------
# Time shift
# ------------------


@pytest.mark.parametrize("shift_samples", [10, -10, 0])
def test_time_shift_left_right(shift_samples):
    n_samples = 2**9
    ir = pf.signals.impulse(n_samples, delay=20)

    ir_truth = pf.signals.impulse(n_samples, 20+shift_samples)
    ir_shifted = dsp.time_shift(ir, shift_samples)

    npt.assert_allclose(ir_shifted.time, ir_truth.time)


def test_time_shift_return_vals():
    n_samples = 2**9
    ir = pf.signals.impulse(n_samples, delay=20)

    ir_shifted = dsp.time_shift(ir, 1, circular_shift=True)
    assert type(ir_shifted) == pf.Signal

    ir_shifted = dsp.time_shift(ir, 1, circular_shift=False)
    assert type(ir_shifted) == pf.TimeData


def test_time_shift_non_circular_left_right():
    shift_samples = 10
    n_samples = 2**9

    ir = pf.signals.impulse(n_samples, delay=20)

    ir_truth = np.zeros(n_samples, dtype=float)
    ir_truth[20+shift_samples] = 1
    ir_truth[:shift_samples] = np.nan

    ir_truth = pf.TimeData(ir_truth, np.arange(n_samples)/ir.sampling_rate)

    ir_shifted = dsp.time_shift(ir, shift_samples, circular_shift=False)
    npt.assert_allclose(ir_shifted.time, ir_truth.time, equal_nan=True)

    shift_samples = -10
    ir_truth = np.zeros(n_samples, dtype=float)
    ir_truth[20+shift_samples] = 1
    ir_truth[shift_samples:] = np.nan
    ir_truth = pf.TimeData(ir_truth, np.arange(n_samples)/ir.sampling_rate)

    ir_shifted = dsp.time_shift(ir, shift_samples, circular_shift=False)
    npt.assert_allclose(ir_shifted.time, ir_truth.time, equal_nan=True)


# ----------------
# Noise power estimation
# ----------------

def test_estimate_noise_power():
    n_samples = 2**18
    rms = 10**(-40/20)
    noise = pf.signals.noise(n_samples, rms=rms)
    actual = dsp.estimate_noise_energy(noise)

    npt.assert_allclose(actual, rms**2, rtol=1e-3, atol=1e-3)


def test_estimate_noise_power_private():
    n_samples = 2**18
    rms = 10**(-40/20)
    noise = pf.signals.noise(n_samples, rms=rms)
    actual = dsp._estimate_noise_energy(noise.time)

    npt.assert_allclose(actual, rms**2, rtol=1e-3, atol=1e-3)


def test_noise_energy_1D():
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=','), 1)
    expected = genfromtxt(
        os.path.join(test_data_path, 'noise_energy_1D.csv'),
        delimiter=',')
    actual = dsp.estimate_noise_energy(
        rir,
        interval=[0.9, 1.0],
        is_energy=False)
    npt.assert_allclose(actual, expected)


def test_noise_energy_2D():
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 1)
    expected = genfromtxt(
        os.path.join(test_data_path, 'noise_energy_2D.csv'),
        delimiter=',')
    actual = dsp.estimate_noise_energy(
        rir,
        interval=[0.9, 1.0],
        is_energy=False)
    npt.assert_allclose(actual, expected)


def test_psnr():
    n_samples = 2**20
    peak_levels = np.array([0, -6, -10])
    noise_level = np.array([-20, -30, -40])
    imp = pf.signals.impulse(n_samples, amplitude=10**(peak_levels/20))
    awgn = pf.signals.noise(n_samples, rms=10**(noise_level/20), seed=7)
    psnr = dsp.peak_signal_to_noise_ratio(imp+awgn)

    npt.assert_allclose(
        1/psnr, 10**((peak_levels + noise_level)/10), rtol=1e-2, atol=1e-2)

# ----------------
# RIR preprocessing
# ----------------

def test_preprocessing_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    rir = pf.Signal(rir, 1)
    actual = dsp.preprocess_rir(
        rir,
        is_energy=False,
        shift=False,
        channel_independent=False)[0]

    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'preprocessing_1D.csv'),
        delimiter=','))
    npt.assert_allclose(actual.time, np.atleast_2d(expected))


def test_preprocessing_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    rir = pf.Signal(rir, 1)

    actual = dsp.preprocess_rir(
        rir,
        is_energy=False,
        shift=False,
        channel_independent=False)

    expected = genfromtxt(
        os.path.join(test_data_path, 'preprocessing_2D.csv'),
        delimiter=',')
    npt.assert_allclose(actual.time, expected)


def test_preprocessing_time_shift_1D(monkeypatch):
    # Patch the RIR start finding to always return same number of samples
    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    rir = pf.Signal(rir, 1)

    actual = dsp.preprocess_rir(
        rir,
        is_energy=False,
        shift=True,
        channel_independent=False)[0]

    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'preprocessing_time_shift_1D.csv'),
        delimiter=','))
    npt.assert_allclose(actual.time, expected)


def test_preprocessing_time_shift_2D(monkeypatch):
    # Patch the RIR start finding to always return same number of samples
    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    rir = pf.Signal(
        genfromtxt(
            os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
            delimiter=','),
        1)

    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'preprocessing_time_shift_2D.csv'),
        delimiter=','))

    actual = dsp.preprocess_rir(
        rir,
        is_energy=False,
        shift=True,
        channel_independent=False)
    npt.assert_allclose(actual.time, expected)


def test_preprocessing_time_shift_channel_independent_1D(monkeypatch):
    # Patch the RIR start finding to always return same number of samples
    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    rir = pf.Signal(
        genfromtxt(
            os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
            delimiter=','),
        1)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(
            test_data_path,
            'preprocessing_time_shift_channel_independent_1D.csv'),
        delimiter=','))

    actual = dsp.preprocess_rir(
        rir,
        is_energy=False,
        shift=True,
        channel_independent=True)[0]
    npt.assert_allclose(actual.time, expected)


def test_preprocessing_time_shift_channel_independent_2D(monkeypatch):
    # Patch the RIR start finding to always return same number of samples
    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 1)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(
            test_data_path,
            'preprocessing_time_shift_channel_independent_2D.csv'),
        delimiter=','))

    actual = dsp.preprocess_rir(
        rir,
        is_energy=False,
        shift=True,
        channel_independent=True)
    npt.assert_allclose(actual.time, expected)


def test_smoothed_rir_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'smoothed_rir_1D.csv'),
        delimiter=',')[np.newaxis]
    actual = dsp._smooth_rir(
        rir,
        sampling_rate=3000,
        smooth_block_length=0.075)[0]
    npt.assert_allclose(actual, expected)


def test_smoothed_rir_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'smoothed_rir_2D.csv'),
        delimiter=',')
    actual = dsp._smooth_rir(
        rir,
        sampling_rate=3000,
        smooth_block_length=0.075)[0]
    npt.assert_allclose(actual, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tests for reverberation time related things. """
from pytest import raises

import numpy as np
import numpy.testing as npt

import pyrato as ra
import pyfar as pf
import pytest


@pytest.mark.parametrize(
    'tx', ['T20', 'T30', 'T40', 'T50', 'T60', 'LDT', 'EDT'])
def test_rt_from_edc(tx):
    times = np.linspace(0, 1.5, 2**9)
    m = -60
    edc = times * m
    edc_exp = pf.TimeData(10**(edc/10), times)
    RT_est = ra.reverberation_time_linear_regression(
        edc_exp, T=tx)
    npt.assert_allclose(RT_est, 1.)


@pytest.mark.parametrize(
    'tx', ['T20', 'T30', 'T40', 'T50', 'T60', 'LDT', 'EDT'])
def test_rt_from_edc_mulitchannel(tx):
    times = np.linspace(0, 1.5, 2**9)
    Ts = np.array([1, 2, 1.5])
    m = -60
    edc = np.atleast_2d(m/Ts).T @ np.atleast_2d(times)
    edc_exp = pf.TimeData(10**(edc/10), times)
    RT_est = ra.reverberation_time_linear_regression(
        edc_exp, T=tx)
    npt.assert_allclose(RT_est, Ts)


@pytest.mark.parametrize(
    'tx', ['T20', 'T30', 'T40', 'T50', 'T60', 'LDT', 'EDT'])
def test_rt_from_edc_mulitchannel_amplitude(tx):
    times = np.linspace(0, 5/2, 2**9)
    Ts = np.array([[1, 2, 1.5], [3, 4, 5]])
    As = np.array([[0, 3, 6], [1, 1, 1]])
    m = -60
    edc = np.zeros((*Ts.shape, times.size))
    for idx in np.ndindex(Ts.shape):
        edc[idx] = As[idx] + m*times/Ts[idx]

    edc_exp = pf.TimeData(10**(edc/10), times)
    RT_est, A_est = ra.reverberation_time_linear_regression(
        edc_exp, T=tx, return_intercept=True)
    npt.assert_allclose(RT_est, Ts)
    npt.assert_allclose(A_est, 10**(As/10))


def test_rt_from_edc_error():
    times = np.linspace(0, 1.5, 2**9)
    m = -60
    edc = times * m
    edc_exp = pf.TimeData(10**(edc/10), times)
    T = 'Bla'

    with raises(ValueError, match='is not a valid interval.'):
        ra.reverberation_time_linear_regression(edc_exp, T=T)



import pytest
import warnings
import pyfar as pf
import pyrato
import numpy as np


def test_warning_fractional_octave_bands():

    with pytest.warns(DeprecationWarning, match='0.5.0'):
        sig = pf.Signal([1, 2, 3], 44100)
        pyrato.dsp.filter_fractional_octave_bands(sig, 1)


def test_warning_center_frequencies_thirds():

    nominal = np.array([
        25, 31.5, 40, 50, 63, 80, 100, 125, 160,
        200, 250, 315, 400, 500, 630, 800, 1000,
        1250, 1600, 2000, 2500, 3150, 4000, 5000,
        6300, 8000, 10000, 12500, 16000, 20000], dtype=float)

    with pytest.warns(DeprecationWarning, match='0.5.0'):
        nom = pyrato.dsp.center_frequencies_third_octaves()[0]

    np.testing.assert_allclose(nom, nominal)


def test_warning_center_frequencies_octaves():
    with pytest.warns(DeprecationWarning, match='0.5.0'):
        nom, exact = pyrato.dsp.center_frequencies_octaves()


def test_warning_start_ir():
    with pytest.warns(DeprecationWarning, match='0.5.0'):

        sig = pf.Signal([0, 0, 1, 0, 0], 44100)
        pyrato.dsp.find_impulse_response_start(sig)


def test_warning_rt_edc():
    with pytest.warns(DeprecationWarning, match='0.5.0'):
        edc = pf.TimeData([-5., -60.], [0.1, 1])
        pyrato.reverberation_time_energy_decay_curve(edc)



# -*- coding: utf-8 -*-

"""Signal processing related functions."""

import warnings

import pyfar as pf
import numpy as np


def find_impulse_response_start(
        impulse_response,
        threshold=20):
    """Find the start sample of an impulse response.
    The start sample is identified as the first sample which is below the
    ``threshold`` level relative to the maximum level of the impulse response.
    For room impulse responses, ISO 3382 [#]_ specifies a threshold of 20 dB.
    This function is primary intended to be used when processing room impulse
    responses.

    Parameters
    ----------
    impulse_response : pyfar.Signal
        The impulse response
    threshold : float, optional
        The threshold level in dB, by default 20, which complies with ISO 3382.
    Returns
    -------
    start_sample : numpy.ndarray, int
        Sample at which the impulse response starts
    Notes
    -----
    The function tries to estimate the PSNR in the IR based on the signal
    power in the last 10 percent of the IR. The automatic estimation may fail
    if the noise spectrum is not white or the impulse response contains
    non-linear distortions. If the PSNR is lower than the specified threshold,
    the function will issue a warning.
    References
    ----------
    .. [#]  ISO 3382-1:2009-10, Acoustics - Measurement of the reverberation
            time of rooms with reference to other acoustical parameters. pp. 22
    Examples
    --------
    Create a band-limited impulse shifted by 0.5 samples and estimate the
    starting sample of the impulse and plot.
    .. plot::
        >>> import pyfar as pf
        >>> import numpy as np
        >>> n_samples = 256
        >>> delay_samples = n_samples // 2 + 1/2
        >>> ir = pf.signals.impulse(n_samples)
        >>> ir = pf.dsp.linear_phase(ir, delay_samples, unit='samples')
        >>> start_samples = pf.dsp.find_impulse_response_start(ir)
        >>> ax = pf.plot.time(ir, unit='ms', label='impulse response', dB=True)
        >>> ax.axvline(
        ...     start_samples/ir.sampling_rate*1e3,
        ...     color='k', linestyle='-.', label='start sample')
        >>> ax.axhline(
        ...     20*np.log10(np.max(np.abs(ir.time)))-20,
        ...     color='k', linestyle=':', label='threshold')
        >>> ax.legend()
    Create a train of weighted impulses with levels below and above the
    threshold, serving as a very abstract room impulse response. The starting
    sample is identified as the last sample below the threshold relative to the
    maximum of the impulse response.
    .. plot::
        >>> import pyfar as pf
        >>> import numpy as np
        >>> n_samples = 64
        >>> delays = np.array([14, 22, 26, 30, 33])
        >>> amplitudes = np.array([-35, -22, -6, 0, -9], dtype=float)
        >>> ir = pf.signals.impulse(n_samples, delays, 10**(amplitudes/20))
        >>> ir.time = np.sum(ir.time, axis=0)
        >>> start_sample_est = pf.dsp.find_impulse_response_start(
        ...     ir, threshold=20)
        >>> ax = pf.plot.time(
        ...     ir, dB=True, unit='samples',
        ...     label=f'peak samples: {delays}')
        >>> ax.axvline(
        ...     start_sample_est, linestyle='-.', color='k',
        ...     label=f'ir start sample: {start_sample_est}')
        >>> ax.axhline(
        ...     20*np.log10(np.max(np.abs(ir.time)))-20,
        ...     color='k', linestyle=':', label='threshold')
        >>> ax.legend()
    """
    warnings.warn(
        "This function will be deprecated in version 0.5.0 "
        "Use pyfar.dsp.find_impulse_response_start instead",
        DeprecationWarning)

    return pf.dsp.find_impulse_response_start(impulse_response, threshold)


def find_impulse_response_maximum(
        impulse_response,
        threshold=20,
        noise_energy='auto'):
    """Find the maximum of an impulse response as argmax(h(t)).
    Performs an initial SNR check according to a defined threshold level in dB.

    Parameters
    ----------
    impulse_response : ndarray, double
        The impulse response
    threshold : double, optional
        Threshold SNR value in dB

    Returns
    -------
    max_sample : int
        Sample at which the impulse response starts

    Note
    ----
    The function tries to estimate the SNR in the IR based on the signal energy
    in the last 10 percent of the IR.

    """
    ir_squared = np.abs(impulse_response.time)**2

    mask_start = int(0.9*ir_squared.shape[-1])
    if noise_energy == 'auto':
        mask = np.arange(mask_start, ir_squared.shape[-1])
        noise = np.mean(np.take(ir_squared, mask, axis=-1), axis=-1)
    else:
        noise = noise_energy

    max_sample = np.argmax(ir_squared, axis=-1)
    max_value = np.max(ir_squared, axis=-1)

    if np.any(max_value < 10**(threshold/10) * noise) or \
            np.any(max_sample > mask_start):
        warnings.warn(
            "The SNR seems lower than the specified threshold value. Check "
            "if this is a valid impulse response with sufficient SNR.")

    return np.squeeze(max_sample)


def time_shift(signal, shift, circular_shift=True, unit='samples'):
    """Apply a time-shift to a signal.

    By default, the shift is performed as a cyclic shift on the time axis,
    potentially resulting in non-causal signals for negative shift values.
    Use the option ``circular_shift=False`` to pad with nan values instead,
    note that in this case the return type will be a ``pyfar.TimeData``.

    Parameters
    ----------
    signal : Signal
        The signal to be shifted
    shift : int, float
        The time-shift value. A positive value will result in right shift on
        the time axis (delaying of the signal), whereas a negative value
        yields a left shift on the time axis (non-causal shift to a earlier
        time). If a single value is given, the same time shift will be applied
        to each channel of the signal. Individual time shifts for each channel
        can be performed by passing an array matching the signals channel
        dimensions ``cshape``.
    unit : str, optional
        Unit of the shift variable, this can be either ``'samples'`` or ``'s'``
        for seconds. By default ``'samples'`` is used. Note that in the case
        of specifying the shift time in seconds, the value is rounded to the
        next integer sample value to perform the shift.
    circular_shift : bool, True
        Perform a circular or non-circular shift. If a non-circular shift is
        performed, the data will be padded with nan values at the respective
        beginning or ending of the data, corresponding to the number of samples
        the data is shifted. In this case, a ``pyfar.TimeData`` object is
        returned.


    Returns
    -------
    pyfar.Signal, pyfar.TimeData
        The time-shifted signal. If a circular shift is performed, the return
        value will be a ``pyfar.Signal``, in case of a non-circular shift, its
        type will be ``pyfar.TimeData``.

    Notes
    -----
    This function is primarily intended to be used when processing room impulse
    responses. When ``circular_shift=True``, the function input is passed into
    ``pyfar.dsp.time_shift``.

    Examples
    --------
    Perform a circular shift a set of ideal impulses stored in three different
    channels and plot the resulting signals

    .. plot::

        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> import matplotlib.pyplot as plt
        >>> impulse = pf.signals.impulse(
        ...     32, amplitude=(1, 1.5, 1), delay=(14, 15, 16))
        >>> shifted = ra.time_shift(impulse, [-2, 0, 2])
        >>> pf.plot.use('light')
        >>> _, axs = plt.subplots(2, 1)
        >>> pf.plot.time(impulse, ax=axs[0])
        >>> pf.plot.time(shifted, ax=axs[1])
        >>> axs[0].set_title('Original signals')
        >>> axs[1].set_title('Shifted signals')
        >>> plt.tight_layout()

    Perform a non-circular shift a single impulse and plot the results.

    .. plot::

        >>> import pyfar as pf
        >>> import pyrato as ra
        >>> import matplotlib.pyplot as plt
        >>> impulse = pf.signals.impulse(32, delay=15)
        >>> shifted = ra.time_shift(impulse, -10, circular_shift=False)
        >>> pf.plot.use('light')
        >>> _, axs = plt.subplots(2, 1)
        >>> pf.plot.time(impulse, ax=axs[0])
        >>> pf.plot.time(shifted, ax=axs[1])
        >>> axs[0].set_title('Original signal')
        >>> axs[1].set_title('Shifted signal')
        >>> plt.tight_layout()

    """
    shift = np.atleast_1d(shift)
    if shift.size == 1:
        shift = np.ones(signal.cshape) * shift

    if unit == 's':
        shift_samples = np.round(shift*signal.sampling_rate).astype(int)
    elif unit == 'samples':
        shift_samples = shift.astype(int)
    else:
        raise ValueError(
            f"Unit is: {unit}, but has to be 'samples' or 's'.")

    shifted = pf.dsp.time_shift(signal, shift_samples, unit='samples')

    if circular_shift is False:
        # Convert to TimeData, as filling with nans will break Fourier trafos
        shifted = pf.TimeData(
            shifted.time,
            shifted.times,
            comment=shifted.comment)

        for ch in np.ndindex(shifted.cshape):
            if shift[ch] < 0:
                shifted.time[ch, shift_samples[ch]:] = np.nan
            else:
                shifted.time[ch, :shift_samples[ch]] = np.nan

    return shifted


def center_frequencies_octaves():
    """Return the octave center frequencies according to the IEC 61260:1:2014
    standard.
    Returns
    -------
    frequencies : ndarray, float
        Octave center frequencies
    """
    warnings.warn(
        "This function will be deprecated in version 0.5.0 "
        "Use pyfar.dsp.filter.fractional_octave_frequencies instead",
        DeprecationWarning)

    nominal, exact = pf.dsp.filter.fractional_octave_frequencies(
        1, (20, 20e3), return_cutoff=False)

    return nominal, exact


def center_frequencies_third_octaves():
    """Return the third octave center frequencies according
    to the ICE 61260:1:2014 standard.
    Returns
    -------
    frequencies : ndarray, float
        third octave center frequencies
    """
    warnings.warn(
        "This function will be deprecated in version 0.5.0 "
        "Use pyfar.dsp.filter.fractional_octave_frequencies instead",
        DeprecationWarning)

    nominal, exact = pf.dsp.filter.fractional_octave_frequencies(
        3, (20, 20e3), return_cutoff=False)

    return nominal, exact


def filter_fractional_octave_bands(
        signal, num_fractions,
        freq_range=(20.0, 20e3), order=6):
    """Apply a fractional octave filter to a signal.
    Filter bank implementation using second order sections of butterworth
    filters for increased numeric accuracy and stability.

    Parameters
    ----------
    signal : ndarray
        input signal to be filtered
    num_fractions : integer
        number of octave fractions
    order : integer, optional
        order of the butterworth filter

    Returns
    -------
    signal_filtered : ndarray
        Signal filtered into fractional octave bands.
    """
    warnings.warn(
        "This function will be deprecated in version 0.5.0 "
        "Use pyfar.dsp.filter.fractional_octave_bands instead",
        DeprecationWarning)

    return pf.dsp.filter.fractional_octave_bands(
        signal, num_fractions, freq_range=freq_range, order=order)


def estimate_noise_energy(
        data,
        interval=[0.9, 1.0],
        is_energy=False):
    """ This function estimates the noise energy level of a given room impulse
    response. The noise is assumed to be Gaussian.

    Parameters
    ----------
    data: np.array
        The room impulse response with dimension [..., n_samples]
    interval : tuple, float, [0.9, 1.]
        Defines the interval of the RIR to be evaluated for the estimation.
        The interval is relative to the length of the RIR [0 = 0%, 1=100%)]
    is_energy: Boolean
        Defines if the data is already squared.

    Returns
    -------
    noise_energy: float
        The energy of the background noise.
    """

    energy_data = preprocess_rir(
        data,
        is_energy=is_energy,
        shift=False,
        channel_independent=False)

    return _estimate_noise_energy(energy_data.time, interval=interval)


def _estimate_noise_energy(
        energy_data,
        interval=[0.9, 1.0]):
    """ This function estimates the noise energy level of a given room impulse
    response. The noise is assumed to be Gaussian.

    Parameters
    ----------
    data: np.array
        The room impulse response with dimension [..., n_samples]
    interval : tuple, float, [0.9, 1.]
        Defines the interval of the RIR to be evaluated for the estimation.
        The interval is relative to the length of the RIR [0 = 0%, 1=100%)]
    is_energy: Boolean
        Defines if the data is already squared.

    Returns
    -------
    noise_energy: float
        The energy of the background noise.
    """

    if np.any(energy_data) < 0:
        raise ValueError("Energy is negative, check your input signal.")

    region_start_idx = int(energy_data.shape[-1]*interval[0])
    region_end_idx = int(energy_data.shape[-1]*interval[1])
    mask = np.arange(region_start_idx, region_end_idx)
    noise_energy = np.nanmean(np.take(energy_data, mask, axis=-1), axis=-1)

    return noise_energy


def _smooth_rir(
        data,
        sampling_rate,
        smooth_block_length=0.075):
    """ Smoothens the RIR by averaging the data in an specified interval.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    sampling_rate: integer
        Defines the sampling rate of the room impulse response.
    smooth_block_length : double
        Defines the block-length of the smoothing algorithm in seconds.

    Returns
    -------
    time_window_data : ndarray, double
        The smoothed RIR.
    time_vector_window : ndarray, double
        The respective time vector fitting the smoothed data.
    time_vector : ndarray, double
        The time vector fitting the original data.

    """
    data = np.atleast_2d(data)
    n_samples = data.shape[-1]
    n_samples_nan = np.count_nonzero(np.isnan(data), axis=-1)

    n_samples_per_block = int(np.round(smooth_block_length * sampling_rate, 0))
    n_blocks = np.asarray(
        np.floor((n_samples-n_samples_nan)/n_samples_per_block),
        dtype=int)

    n_blocks_min = int(np.min(n_blocks))
    n_samples_actual = int(n_blocks_min*n_samples_per_block)
    reshaped_array = np.reshape(
        data[..., :n_samples_actual],
        (-1, n_blocks_min, n_samples_per_block))
    time_window_data = np.mean(reshaped_array, axis=-1)

    # Use average time instances corresponding to the average energy level
    # instead of time for the first sample of the block
    time_vector_window = \
        ((0.5+np.arange(0, n_blocks_min)) * n_samples_per_block/sampling_rate)

    # Use the time corresponding to the sampling of the original data
    time_vector = (np.arange(0, n_samples))/sampling_rate

    return time_window_data, time_vector_window, time_vector


def preprocess_rir(
        data,
        is_energy=False,
        shift=False,
        channel_independent=False):
    """ Preprocess the room impulse response for further processing:
        - Square data
        - Shift the RIR to the first sample of the array, compensating for the
          delay of the time of arrival of the direct sound. The time shift is
          performed as a non-cyclic shift, adding numpy.nan values in the end
          of the RIR corresponding to the number of samples the data is
          shifted by.
        - The time shift can be done channel-independent or not.

    Parameters
    ----------
    data : ndarray, double
        The room impulse response with dimension [..., n_samples]
    is_energy : boolean
        Defines, if the data is already squared.
    shift : boolean
        Defines, if the silence at beginning of the RIR should be removed.
    channel_independent : boolean
        Defines, if the time shift is done channel-independent or not.

    Returns
    -------
    energy_data : ndarray, double
        The preprocessed RIR

    """
    times = data.times
    n_channels = np.prod(data.cshape)

    if shift:
        rir_start_idx = find_impulse_response_start(data)

        if channel_independent and not n_channels == 1:
            shift_samples = -rir_start_idx
        else:
            min_shift = np.amin(rir_start_idx)
            shift_samples = np.asarray(
                -min_shift * np.ones(data.cshape), dtype=int)

        result = time_shift(
            data, shift_samples, circular_shift=False)
    else:
        result = data

    if not is_energy:
        energy_data = np.abs(result.time)**2
    else:
        energy_data = result.time.copy()

    energy_data = pf.TimeData(energy_data, times)

    return energy_data


def peak_signal_to_noise_ratio(
        impulse_response,
        noise_power='auto',
        is_energy=False):
    """Calculate the peak-signal-to-noise-ratio of an impulse response.

    Parameters
    ----------
    impulse_response : pyfar.Signal
        The impulse response
    noise_power : float, str, optional
        The noise power. The default is 'auto', in which case the noise power
        is estimated from the last 10 % of the impulse response.

    Returns
    -------
    float, array-like
        The estimated peak-signal-to-noise-ratio for each channel of the
        impulse response.

    """
    data = impulse_response.time
    if is_energy is False:
        data = data**2

    if noise_power == 'auto':
        noise_power = _estimate_noise_energy(data)

    return np.max(data, axis=-1) / noise_power



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pyfar>=0.5.0',
    'numpy>=1.14.0',
    'scipy>=1.5.0',
    'matplotlib']

setup_requirements = ['pytest-runner', ]

test_requirements = [
    'pytest',
    'bump2version',
    'wheel',
    'watchdog',
    'flake8',
    'tox',
    'coverage',
    'Sphinx',
    'twine'
]

setup(
    author="The pyfar developers",
    author_email='info@pyfar.org',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    description="Collection of functions commonly used in room acoustics",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    include_package_data=True,
    keywords='pyrato',
    name='pyrato',
    packages=find_packages(include=['pyrato']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/pyfar/pyrato',
    version='0.4.0',
    zip_safe=False,
    python_requires='>=3.8'
)



"""Room acoustic parameters
"""
import re
import numpy as np


def reverberation_time_linear_regression(
        energy_decay_curve, T='T20', return_intercept=False):
    """Estimate the reverberation time from a given energy decay curve.
    The linear regression is performed using least squares error minimization
    according to the ISO standard 3382 [#]_.

    Parameters
    ----------
    energy_decay_curve : ndarray, double
        Energy decay curve. The time needs to be the arrays last dimension.
    times : ndarray, double
        Time vector corresponding to each sample of the EDC.
    T : 'T20', 'T30', 'T40', 'T50', 'T60', 'EDT', 'LDT'
        Decay interval to be used for the reverberation time extrapolation. EDT
        corresponds to the early decay time extrapolated from the interval
        [0, -10] dB, LDT corresponds to the late decay time extrapolated from
        the interval [-25, -35] dB.
    normalize : bool, True
        Normalize the EDC to the steady state energy level

    Returns
    -------
    reverberation_time : double
        The reverberation time

    References
    ----------
    .. [#]  ISO 3382, Acoustics - Measurement of the reverberation time of
            rooms with reference to other acoustical parameters.

    Examples
    --------

    Estimate the reverberation time from an energy decay curve.

    >>> import numpy as np
    >>> import pyfar as pf
    >>> import pyrato as ra
    >>> from pyrato.analytic import rectangular_room_rigid_walls
    ...
    >>> L = np.array([8, 5, 3])/10
    >>> source_pos = np.array([5, 3, 1.2])/10
    >>> receiver_pos = np.array([1, 1, 1.2])/10
    >>> rir, _ = rectangular_room_rigid_walls(
    ...     L, source_pos, receiver_pos,
    ...     reverberation_time=1, max_freq=1.5e3, n_samples=2**12,
    ...     speed_of_sound=343.9, samplingrate=3e3)
    >>> rir = rir/rir.time.max()
    ...
    >>> awgn = pf.signals.noise(
    ...     rir.n_samples, rms=10**(-50/20),
    ...     sampling_rate=rir.sampling_rate)
    >>> rir = rir + awgn
    ...
    >>> edc = ra.energy_decay_curve_chu_lundeby(rir)
    >>> t_20 = ra.reverberation_time_linear_regression(edc, 'T20')
    >>> t_20
    ...     array([0.99526253])

    """
    intervals = [20, 30, 40, 50, 60]

    if T == 'EDT':
        upper = -0.1
        lower = -10.1
    elif T == 'LDT':
        upper = -25.
        lower = -35.
    else:
        try:
            (int(re.findall(r'\d+', T)[0]) in intervals)
        except IndexError:
            raise ValueError(
                "{} is not a valid interval for the regression.".format(T))

        upper = -5
        lower = -np.double(re.findall(r'\d+', T)) + upper

    edcs_db = 10*np.log10(np.abs(energy_decay_curve.time))
    times = energy_decay_curve.times

    reverberation_times = np.zeros(energy_decay_curve.cshape, dtype=float)
    intercepts = np.zeros(energy_decay_curve.cshape, dtype=float)

    for ch in np.ndindex(energy_decay_curve.cshape):
        edc_db = edcs_db[ch]
        idx_upper = np.nanargmin(np.abs(upper - edc_db))
        idx_lower = np.nanargmin(np.abs(lower - edc_db))

        A = np.vstack(
            [times[idx_upper:idx_lower], np.ones(idx_lower - idx_upper)]).T
        gradient, const = np.linalg.lstsq(
            A, edc_db[..., idx_upper:idx_lower], rcond=None)[0]

        reverberation_times[ch] = -60 / gradient
        intercepts[ch] = 10**(const/10)

    if return_intercept is True:
        return reverberation_times, intercepts
    else:
        return reverberation_times



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    The test_edc_noise_handling module provides the functionality to test
    the functions of the module edc_noise_handling.
"""

import numpy as np
import os
import numpy.testing as npt
from pyrato import edc as enh
from numpy import genfromtxt
import pyfar as pf
test_data_path = os.path.join(os.path.dirname(__file__), 'test_data')


def mock_shift_samples_1d(*args, **kwargs):
    return np.array([76])


def mock_shift_samples_2d(*args, **kwargs):
    return np.array([76, 76])


def test_substracted_1D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'substracted_1D.csv'),
        delimiter=',')
    actual = enh._subtract_noise_from_squared_rir(rir**2)
    npt.assert_allclose(actual, expected)


def test_substracted_2D():
    rir = genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=',')
    expected = genfromtxt(
        os.path.join(test_data_path, 'substracted_2D.csv'),
        delimiter=',')
    actual = enh._subtract_noise_from_squared_rir(rir**2)
    npt.assert_allclose(actual, expected)


def test_edc_truncation_1D():
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_truncation_1D.csv'),
        delimiter=','))

    actual = enh.energy_decay_curve_truncation(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=-np.inf)

    pf.plot.time(actual, dB=True, log_prefix=10)

    npt.assert_allclose(actual.time, expected)

    actual = enh.energy_decay_curve_truncation(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=15)

    mask = expected < 10**((-40+15)/10)
    expected[mask] = np.nan

    pf.plot.time(actual, dB=True, log_prefix=10)

    npt.assert_allclose(actual.time, expected)


def test_edc_truncation_2D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_truncation_2D.csv'),
        delimiter=','))

    actual = enh.energy_decay_curve_truncation(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=-np.inf)
    npt.assert_allclose(actual.time, expected)


def test_edc_lundeby_1D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_1D.csv'),
        delimiter=','))

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.energy_decay_curve_lundeby(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual.time, expected)


def test_edc_lundeby_2D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_2D.csv'),
        delimiter=','))

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_lundeby(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual.time, expected)


def test_edc_lundeby_chu_1D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_chu_1D.csv'),
        delimiter=','))

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.energy_decay_curve_chu_lundeby(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual.time, expected)


def test_edc_lundeby_chu_2D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_lundeby_chu_2D.csv'),
        delimiter=','))

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_chu_lundeby(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        plot=False)
    npt.assert_allclose(actual.time, expected)


def test_edc_chu_1D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_chu_1D.csv'),
        delimiter=','))

    actual = enh.energy_decay_curve_chu(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=None,
        plot=False)
    npt.assert_allclose(actual.time, expected)

    # Test with a sufficiently high threshold to ensure exact matching of nans
    threshold = 15
    actual = enh.energy_decay_curve_chu(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=threshold,
        plot=False)

    mask = expected <= 10**((-40+threshold)/10)
    expected[mask] = np.nan

    pf.plot.time(actual, dB=True, log_prefix=10)

    pf.plot.time(pf.TimeData(expected, actual.times), dB=True, log_prefix=10)
    npt.assert_allclose(actual.time, expected)


def test_edc_chu_2D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 3e3)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'edc_chu_2D.csv'),
        delimiter=','))

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.energy_decay_curve_chu(
        rir,
        is_energy=False,
        time_shift=True,
        channel_independent=False,
        normalize=True,
        threshold=None,
        plot=False)
    npt.assert_allclose(actual.time, expected)


def test_intersection_time_1D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_1D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'intersection_time_1D.csv'),
        delimiter=',')).T

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_1d)

    actual = enh.intersection_time_lundeby(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=False,
        channel_independent=False,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_intersection_time_2D(monkeypatch):
    rir = pf.Signal(genfromtxt(
        os.path.join(test_data_path, 'analytic_rir_psnr50_2D.csv'),
        delimiter=','), 3000)
    expected = np.atleast_2d(genfromtxt(
        os.path.join(test_data_path, 'intersection_time_2D.csv'),
        delimiter=','))

    # monkeypatch.setattr(
    #     dsp,
    #     "find_impulse_response_start",
    #     mock_shift_samples_2d)

    actual = enh.intersection_time_lundeby(
        rir,
        freq='broadband',
        is_energy=False,
        time_shift=False,
        channel_independent=False,
        plot=False)
    npt.assert_allclose(actual, expected)


def test_energy_decay_curve_threshold():

    t_60 = 1
    m = -60/t_60

    n_samples = 10
    times = np.linspace(0, 1, n_samples)
    edc_log = np.atleast_2d(times * m)

    edc_log = np.tile(edc_log, (2, 3, 1))

    edc = enh._truncate_energy_decay_curve(10**(edc_log.copy()/10), 30)

    edc_ref = 10**(edc_log.copy()/10)
    edc_ref[..., n_samples//2:] = np.nan

    npt.assert_allclose(edc, edc_ref)


def test_truncate_energy_decay_curve():
    t_60 = 1
    m = -60/t_60

    n_samples = 10
    times = np.linspace(0, 1, n_samples)
    edc_log = np.atleast_2d(times * m)

    edc_log = np.tile(edc_log, (2, 3, 1))

    edc = pf.TimeData(10**(edc_log.copy()/10), times)
    edc_trunc = enh.truncate_energy_decay_curve(edc, 30)

    edc_ref = 10**(edc_log.copy()/10)
    edc_ref[..., n_samples//2:] = np.nan

    npt.assert_allclose(edc_trunc.time, edc_ref)



# -*- coding: utf-8 -*-

__author__ = \
    """The pyfar developers"""
__email__ = 'info@pyfar.org'
__version__ = '0.4.0'

from .rap import (
    reverberation_time_linear_regression
)
from .roomacoustics import (
    reverberation_time_energy_decay_curve,
    energy_decay_curve_analytic,
    air_attenuation_coefficient,
)
from .dsp import (
    find_impulse_response_maximum,
    find_impulse_response_start,
    time_shift,
    preprocess_rir,
    estimate_noise_energy,
)
from .edc import (
    schroeder_integration,
    energy_decay_curve_chu,
    energy_decay_curve_chu_lundeby,
    energy_decay_curve_lundeby,
    energy_decay_curve_truncation,
    intersection_time_lundeby,
)

__all__ = [
    'reverberation_time_linear_regression',
    'reverberation_time_energy_decay_curve',
    'schroeder_integration',
    'energy_decay_curve_analytic',
    'air_attenuation_coefficient',
    'find_impulse_response_maximum',
    'find_impulse_response_start',
    'time_shift',
    'preprocess_rir',
    'energy_decay_curve_chu',
    'energy_decay_curve_chu_lundeby',
    'energy_decay_curve_lundeby',
    'energy_decay_curve_truncation',
    'estimate_noise_energy',
    'intersection_time_lundeby',
]
